import numpy as np
from scipy.sparse import lil_matrix, csc_matrix
import combinatorics
import multiindex
import f90_simulate


def simulated_parameters(n=10, m=3, degs=[1], dist=None, dist_par=None):

    # helper variables
    mn = m * n
    mm = m * np.arange(n)

    if (dist is None) or (dist_par is None):
        dist = np.random.normal
        dist_par = (0.0, 1.0 / np.sqrt(mn))

    degs = np.array(degs)
    # monomials with degree larger than n guaranteed 0
    degs = degs[degs <= n]
    # degrees greater than 0
    degs1 = degs[degs > 0]
    # degrees greater than 1
    degs2 = degs[degs > 1]
    max_deg = degs.max()

    p = combinatorics.binomial_coefficients(n, max_deg)
    q = m**np.arange(max_deg + 1)

    # model parameters
    f = {deg: dist(*dist_par, size=(mn, p[deg] * q[deg])) for deg in degs}
    if 0 in degs:
        f[0] = np.vstack([fi - fi.mean(0) for fi in np.split(f[0], n)])
    if 1 in degs:
        f[1] = np.vstack([fi - fi.mean(0) for fi in np.split(f[1], n)])
        f[1] = np.hstack([
            fj - fj.mean(1)[:, np.newaxis] for fj in np.split(f[1], n, axis=1)
        ])

    return f


def simulated_data(par, n, m, l=None, o=1.0):

    degs = np.sort(par.keys())
    k = degs.shape[0]

    max_deg = degs.max()
    p = combinatorics.binomial_coefficients(n, max_deg)[degs]
    q = m**np.arange(max_deg + 1)[degs]
    pq = p * q

    p_sum = p.sum()

    x_idx = [multiindex.get_multiindex(n=n, m=m, deg=deg) for deg in degs]
    x_idx = 1 + np.hstack([i.flatten() for i in x_idx])

    par = np.hstack([par[deg] for deg in degs])

    if l is None:
        l = int(o * np.prod(par.shape))

    return f90_simulate.simulate_data(
        par,
        degs,
        p,
        q,
        x_idx,
        n,
        m,
        p_sum,
        l,
        nm=n * m,
        pq_sum=(p * q).sum(),
        k=k,
        pdegs_sum=(p * degs).sum())



import combinatorics
import multiprocessing as mp
from scipy.sparse.linalg import svds
import numpy as np
import f90_fit
from scipy.sparse import csc_matrix
import multiprocessing as mp


def fit(x, y, m_x, m_y, degs, max_iters=100, iters=None, return_all=True):
    # x: sum(p) by l
    # ------------------------------------
    # x1: x[i_x[0]:i_x[1], :] -- p[0] by l
    # ------------------------------------
    # x2: x[i_x[1]:i_x[2], :] -- p[1] by l
    # ------------------------------------
    # ...
    # ------------------------------------
    # i_x = np.insert(p.cumsum(), 0, 0)

    if x.shape[-1] != y.shape[-1]:
        print 'number of samples must be same for both x and y'
        return -1
    else:
        l = x.shape[1]

    n_x = x.shape[0] if len(x.shape) > 1 else 1
    n_y = y.shape[0] if len(y.shape) > 1 else 1

    ohx = one_hot(x, m=m_x, n=n_x, degs=degs)
    ohy = one_hot(y, m=m_y, n=n_y, degs=degs)

    degs = np.array(degs)
    k = degs.shape[0]

    max_deg = degs.max()
    p = combinatorics.binomial_coefficients(n_x, max_deg)[degs]
    q = m_x**np.arange(max_deg + 1)[degs]
    pq = p * q

    p_sum = p.sum()
    pq_sum = pq.sum()

    ohx_rank = np.linalg.matrix_rank(ohx.todense())
    ohx_svd = svds(ohx, k=ohx_rank)

    ohx_pinv = [ohx_svd[2].T, 1.0 / ohx_svd[1], ohx_svd[0].T]

    iters = 20

    par, disc = [], []

    for yi in y:

        res = f90_fit.fit(
            x,
            yi,
            ohx_pinv[0],
            ohx_pinv[1],
            ohx_pinv[2],
            q,
            p,
            m_y,
            max_iters,
            p_sum=p_sum,
            l=l,
            ohx_rank=ohx_rank,
            pq_sum=pq_sum,
            k=k,
            iters=iters)

        it = res[0]

        if return_all:
            par.append(res[1][:it - 1])
            disc.append(res[2][1:it - 1])
        else:
            par.append(res[1][it - 2].squeeze())
            disc.append(res[2][it - 2])

    return par, disc


def one_hot(x, n, m, degs):
    l = x.shape[-1]

    degs = np.array(degs)
    k = degs.shape[0]

    max_deg = degs.max()
    p = combinatorics.binomial_coefficients(n, max_deg)[degs]
    q = m**np.arange(max_deg + 1)[degs]
    pq = p * q

    p_sum = p.sum()

    i_ohx = np.insert(pq.cumsum(), 0, 0)

    data = np.ones(p_sum * l)
    stratifier = np.hstack(
        [i_ohx[i] + q[i] * np.arange(p[i]) for i in range(k)])
    indices = (x + stratifier[:, np.newaxis]).T.flatten() - 1
    indptr = p_sum * np.arange(l + 1)

    return csc_matrix((data, indices, indptr))



import numpy as np
import combinatorics

def get_multiindex(m, n, deg):

    mn = m * n

    if deg > n:
        return

    p = combinatorics.binomial_coefficients(n, deg)[-1]
    q = m**deg

    vars = np.empty((p, deg), dtype=int)
    var = np.arange(deg)
    vars[0] = var.copy()

    for i in range(1, p):

        idx = deg - 1
        var[idx] += 1
        while var[idx] + deg - idx > n:
            idx -= 1
            var[idx] += 1
        for idx in range(idx + 1, deg):
            var[idx] = var[idx - 1] + 1

        vars[i] = var.copy()

    return vars



from numpy.distutils.core import Extension

fortran_src = ['simulate', 'fit']
fortran_ext = [
    Extension(
        name='f90_' + f,
        sources=['./fem/f90_' + f + '.f90', './fem/f90_' + f + '.pyf'])
    for f in fortran_src
]

with open('README.rst', 'r') as f:
    readme = f.read()

with open('version', 'r') as f:
    version = f.read()

if __name__ == '__main__':
    from numpy.distutils.core import setup

    setup(
        name='fem',
        version=version,
        description='Free Energy Minimization',
        long_description=readme,
        author='Joseph P. McKenna',
        author_email='joepatmckenna@gmail.com',
        url='http://lbm.niddk.nih.gov/mckennajp/fem',
        download_url='https://pypi.org/project/fem',
        packages=['fem'],
        ext_modules=fortran_ext,
        classifiers=("Programming Language :: Python :: 2",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent"),
        license='MIT',
        keywords=['inference', 'statistics', 'machine learning'],
        data_files=[])
